#!/usr/bin/env python
#coding: iso-8859-15
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id$
"""
Daemon that listens to specific OTDB status changes, requests the parset of such jobs including their predecessors, and
posts them on the bus.
"""

from lofar.messaging import FromBus, ToBus, EventMessage # RPC,
from lofar.parameterset import parameterset
from lofar.sas.otdb.OTDBBusListener import OTDBBusListener
from lofar.common.util import waitForInterrupt
from lofar.sas.otdb.otdbrpc import OTDBRPC
from lofar.sas.otdb.config import DEFAULT_OTDB_NOTIFICATION_BUSNAME, DEFAULT_OTDB_NOTIFICATION_SUBJECT
from lofar.sas.resourceassignment.rataskspecified.config import DEFAULT_RA_TASK_SPECIFIED_NOTIFICATION_BUSNAME
from lofar.sas.resourceassignment.rataskspecified.config import DEFAULT_RA_TASK_SPECIFIED_NOTIFICATION_SUBJECT
from lofar.sas.otdb.config import DEFAULT_OTDB_SERVICE_BUSNAME, DEFAULT_OTDB_SERVICENAME

import logging
logger = logging.getLogger(__name__)

""" Prefix that is common to all parset keys, depending on the exact source. """
PARSET_PREFIX="ObsSW."

class RATaskSpecified(OTDBBusListener):
    # Translation table for task types: OTDB (type,subtype) -> RADB (type,subtype)
    OTDB_to_RADB_type_translation_table = {
      ("Observation", "Beam Observation"):    ("observation", "bfmeasurement"),
      ("Observation", "Interferometer"):      ("observation", "interferometer"),
      ("Observation", "TBB (piggyback)"):     ("observation", "tbbmeasurement"),
      ("Observation", "TBB (standalone)"):    ("observation", "tbbmeasurement"),
      ("Pipeline", "Averaging Pipeline"):     ("pipeline", "averaging pipeline"),
      ("Pipeline", "Calibration Pipeline"):   ("pipeline", "calibration pipeline"),
      ("Pipeline", "Imaging Pipeline"):       ("pipeline", "imaging pipeline"),
      ("Pipeline", "Imaging Pipeline MSSS"):  ("pipeline", "imaging pipeline msss"),
      ("Pipeline", "Long Baseline Pipeline"): ("pipeline", "long baseline pipeline"),
      ("Pipeline", "Pulsar Pipeline"):        ("pipeline", "pulsar pipeline"),
      ("MAINTENANCE", ""):                    ("reservation", "maintenance"),
      ("RESERVATION", ""):                    ("reservation", "project"),
    }

    def __init__(self,
                   otdb_notification_busname=DEFAULT_OTDB_NOTIFICATION_BUSNAME,
                   otdb_notification_subject=DEFAULT_OTDB_NOTIFICATION_SUBJECT,
                   otdb_service_busname=DEFAULT_OTDB_SERVICE_BUSNAME,
                   otdb_service_subject=DEFAULT_OTDB_SERVICENAME,
                   notification_busname=DEFAULT_RA_TASK_SPECIFIED_NOTIFICATION_BUSNAME,
                   notification_subject=DEFAULT_RA_TASK_SPECIFIED_NOTIFICATION_SUBJECT,
                   broker=None, **kwargs):
        super(RATaskSpecified, self).__init__(busname=otdb_notification_busname, subject=otdb_notification_subject, **kwargs)
        self.otdbrpc  = OTDBRPC(busname=otdb_service_busname, servicename=otdb_service_subject, broker=broker) ## , ForwardExceptions=True hardcoded in RPCWrapper right now
        self.send_bus = ToBus("%s/%s" % (notification_busname, notification_subject))

    def start_listening(self, **kwargs):
        self.otdbrpc.open()
        self.send_bus.open()
        super(RATaskSpecified, self).start_listening(**kwargs)

    def stop_listening(self, **kwargs):
        super(RATaskSpecified, self).stop_listening(**kwargs)
        self.send_bus.close()
        self.otdbrpc.close()

    @staticmethod
    def resourceIndicatorsFromParset(radb_type, radb_subtype, parset):
        """ Extract the parset keys that are required for resource assignment. """

        subset = {}

        def add(key, conversion=parameterset.getString, prefix=PARSET_PREFIX):
            """ Add the given key to our subset selection, using an optional
                conversion. """
            if parset.isDefined(prefix + key):
                subset[key] = conversion(parset, prefix + key)

        """ Some conversion functions for common parameter-value types."""
        def as_strvector(parset, key):
            return parset.getStringVector(key, True)

        def as_intvector(parset, key):
            return parset.getIntVector(key, True)

        def as_bool(parset, key):
            return parset.getBool(key)

        def as_int(parset, key):
            return parset.getInt(key)

        def as_float(parset, key):
            return parset.getDouble(key)

        # =====================================
        # Parset meta info
        # =====================================
        add("Version.number", as_int, prefix="")

        # =====================================
        # Maintenance/reservation settings
        # =====================================
        if radb_type in ["maintenance","reservation"]:
            add("Observation.VirtualInstrument.stationList", as_strvector)
            add("Observation.startTime")
            add("Observation.stopTime")

        # =====================================
        # Task settings
        # =====================================
        add("Observation.momID")
        add("Observation.startTime")
        add("Observation.stopTime")

        # =====================================
        # Observation settings
        # =====================================
        if radb_type == "observation":
            add("Observation.sampleClock", as_int)
            add("Observation.nrBitsPerSample", as_int)
            add("Observation.antennaSet")
            add("Observation.VirtualInstrument.stationList", as_strvector)
            add("Observation.nrBeams", as_int)

            nrSAPs = subset.get("Observation.nrBeams", 0)
            for sap in xrange(0, nrSAPs):
                add("Observation.Beam[%d].subbandList" % (sap,), as_intvector)

            # =====================================
            # Correlator settings
            # =====================================
            add("Observation.DataProducts.Output_Correlated.enabled", as_bool)
            add("Observation.DataProducts.Output_Correlated.storageClusterName")
            add("Observation.DataProducts.Output_Correlated.identifications", as_strvector)
            add("Observation.ObservationControl.OnlineControl.Cobalt.Correlator.integrationTime", as_float)
            add("Observation.ObservationControl.OnlineControl.Cobalt.Correlator.nrChannelsPerSubband", as_int)

            # =====================================
            # Beamformer settings
            # =====================================
            add("Observation.DataProducts.Output_IncoherentStokes.enabled", as_bool)
            add("Observation.DataProducts.Output_IncoherentStokes.storageClusterName")
            add("Observation.DataProducts.Output_IncoherentStokes.identifications", as_strvector)
            add("Observation.DataProducts.Output_CoherentStokes.enabled", as_bool)
            add("Observation.DataProducts.Output_CoherentStokes.storageClusterName")
            add("Observation.DataProducts.Output_CoherentStokes.identifications", as_strvector)
            add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.flysEye", as_bool)
            #add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.CoherentStokes.nrChannelsPerSubband") # only needed to determine Cobalt.blockSize
            add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.CoherentStokes.subbandsPerFile", as_int)
            add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.CoherentStokes.timeIntegrationFactor", as_int)
            add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.CoherentStokes.which")
            #add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.IncoherentStokes.nrChannelsPerSubband") # only needed to determine Cobalt.blockSize
            add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.IncoherentStokes.subbandsPerFile", as_int)
            add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.IncoherentStokes.timeIntegrationFactor", as_int)
            add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.IncoherentStokes.which")
            for sap in xrange(0, nrSAPs):
                add("Observation.Beam[%d].nrTabRings" % (sap,), as_int)

                add("Observation.Beam[%d].nrTiedArrayBeams" % (sap,), as_int)
                nrTABs = subset.get("Observation.Beam[%d].nrTiedArrayBeams" % (sap,), 0)
                for tab in xrange(0, nrTABs):
                    add("Observation.Beam[%d].TiedArrayBeam[%d].coherent" % (sap,tab), as_bool)

        # =====================================
        # Pipeline settings
        # =====================================
        if radb_type == "pipeline" and radb_subtype in ["averaging pipeline", "calibration pipeline"]:
            # Calibrator / Averaging pipelines
            add("Observation.DataProducts.Output_Correlated.enabled", as_bool)
            add("Observation.DataProducts.Output_Correlated.storageClusterName")
            add("Observation.DataProducts.Output_Correlated.identifications", as_strvector)
            add("Observation.DataProducts.Output_InstrumentModel.enabled", as_bool)
            add("Observation.DataProducts.Output_InstrumentModel.storageClusterName")
            add("Observation.DataProducts.Output_InstrumentModel.identifications", as_strvector)
            add("Observation.DataProducts.Input_Correlated.enabled", as_bool)
            add("Observation.DataProducts.Input_Correlated.identifications", as_strvector)
            add("Observation.DataProducts.Input_InstrumentModel.enabled", as_bool)
            add("Observation.DataProducts.Input_InstrumentModel.identifications", as_strvector)
            add("Observation.ObservationControl.PythonControl.DPPP.demixer.freqstep", as_int)
            add("Observation.ObservationControl.PythonControl.DPPP.demixer.timestep", as_int)

        if radb_type == "pipeline" and radb_subtype in ["imaging pipeline", "imaging pipelines msss"]:
            # Imaging pipeline
            add("Observation.DataProducts.Output_SkyImage.enabled", as_bool)
            add("Observation.DataProducts.Output_SkyImage.storageClusterName")
            add("Observation.DataProducts.Output_SkyImage.identifications", as_strvector)
            add("Observation.DataProducts.Input_Correlated.enabled", as_bool)
            add("Observation.DataProducts.Input_Correlated.identifications", as_strvector)
            add("Observation.ObservationControl.PythonControl.Imaging.slices_per_image")
            add("Observation.ObservationControl.PythonControl.Imaging.subbands_per_image")

        if radb_type == "pipeline" and radb_subtype == "long baseline pipeline":
            # Long-baseline pipeline
            add("Observation.DataProducts.Output_Correlated.enabled", as_bool)
            add("Observation.DataProducts.Output_Correlated.storageClusterName")
            add("Observation.DataProducts.Output_Correlated.identifications", as_strvector)
            add("Observation.DataProducts.Input_Correlated.enabled", as_bool)
            add("Observation.DataProducts.Input_Correlated.identifications", as_strvector)
            add("Observation.ObservationControl.PythonControl.LongBaseline.subbandgroups_per_ms", as_int)
            add("Observation.ObservationControl.PythonControl.LongBaseline.subbands_per_subbandgroup", as_int)

        if radb_type == "pipeline" and radb_subtype == "pulsar pipeline":
            # Pulsar pipeline
            add("Observation.DataProducts.Output_Pulsar.enabled", as_bool)
            add("Observation.DataProducts.Output_Pulsar.storageClusterName")
            add("Observation.DataProducts.Output_Pulsar.identifications", as_strvector)
            add("Observation.DataProducts.Input_CoherentStokes.enabled", as_bool)
            add("Observation.DataProducts.Input_CoherentStokes.identifications", as_strvector)
            add("Observation.DataProducts.Input_IncoherentStokes.enabled", as_bool)
            add("Observation.DataProducts.Input_IncoherentStokes.identifications", as_strvector)

        return subset

    @staticmethod
    def get_predecessors(parset):
        """ Extract the list of predecessor obs IDs from the given parset. """

        predecessors = parset.getStringVector(PARSET_PREFIX + "Observation.Scheduler.predecessors", True)

        # Key contains values starting with 'S' = Scheduler, 'L'/'T' = OTDB, 'M' = MoM
        # 'S' we can probably ignore? Might be only internal in the Scheduler?
        result = []
        for p in predecessors:
            try: # Made the source a string for readability, but it's not efficient
                if p.startswith('M'):
                    result.append({'source': 'mom', 'id': int(p[1:])})
                elif p.startswith('L') or s.startswith('T'):
                    result.append({'source': 'otdb', 'id': int(p[1:])})
                else: # 'S'
                    logger.info("found a predecessor ID I can't handle: %s", p)
                    result.append({'source': 'other', 'id': int(p[1:])})
            except ValueError:
                logger.warning("found a predecessor ID that I can't parse %s", p)
        return result

    @staticmethod
    def get_specification(parset):
        result = {}

        # parse & marshall task type information
        otdb_task_type    = parset.getString(PARSET_PREFIX + "Observation.processType")
        otdb_task_subtype = parset.getString(PARSET_PREFIX + "Observation.processSubtype")
        result['task_type'],result['task_subtype'] = RATaskSpecified.OTDB_to_RADB_type_translation_table[(otdb_task_type,otdb_task_subtype)]

        # parse & marshall specification details
        result['specification'] = RATaskSpecified.resourceIndicatorsFromParset(result['task_type'], result['task_subtype'], parset)

        return result

    def get_specification_with_predecessors(self, id, id_source, state, found_parsets):
        logger.info("Processing ID %s from %s" % (id, id_source))
        if id_source == "other":
            return None
        elif id_source == "mom":
            otdb_id = self.otdbrpc.taskGetIDs( mom_id=id )['otdb_id']
        elif id_source == "otdb":
            otdb_id = id
        else:
            raise ValueError("Error in understanding id_source %s (id %s)" % (id_source, id))

        logger.info("Processing OTDB ID %s", otdb_id)
        result = {"otdb_id": otdb_id, "predecessors": []}
        if state:
            result["state"] = state # TODO should be status not state
        else:
            pass #otdbrpc.taskGetStatus not implemented and maybe not needed?

        if otdb_id in found_parsets:
            parset = found_parsets[otdb_id]
        else:
            # obtain parset & convert dict to parameterset object
            parset = parameterset(self.otdbrpc.taskGetSpecification( otdb_id=otdb_id )['specification'])
            # add it to our cache
            found_parsets[otdb_id] = parset

        logger.info("parset [%s]: %s" % (otdb_id, parset))

        # Fetsh specifications for this parset
        specs = self.get_specification(parset)
        result.update(specs)

        # TODO probably we only need to do this on state=prescheduled, but then we need a different name for the function?
        logger.info("Processing predecessors")
        predecessor_ids = self.get_predecessors(parset)
        for id in predecessor_ids:
            predecessor_result = self.get_specification_with_predecessors(id['id'], id['source'], "", found_parsets)
            if predecessor_result:
                result["predecessors"].append(predecessor_result)
        return result

    def onObservationApproved(self, main_id, modificationTime):
        self.createAndSendSpecifiedTask(main_id, 'approved')

    def onObservationPrescheduled(self, main_id, modificationTime):
        self.createAndSendSpecifiedTask(main_id, 'prescheduled')

    def createAndSendSpecifiedTask(self, main_id, status):
        # Construct root node of tree
        resultTree = self.get_specification_with_predecessors(main_id, "otdb", status, {})
        logger.info("Sending result: %s" % resultTree)

        # Put result on bus
        msg = EventMessage(content=resultTree)
        self.send_bus.send(msg)
        logger.info("Result sent")

def main():
    # make sure we run in UTC timezone
    import os
    os.environ['TZ'] = 'UTC'

    import logging
    import sys
    from optparse import OptionParser
    from lofar.common.util import waitForInterrupt

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser("%prog [options]",
                          description="run the rataskspecified service")
    parser.add_option("-b", "--notification_bus", dest="notification_bus", type="string",
                      default=DEFAULT_RA_TASK_SPECIFIED_NOTIFICATION_BUSNAME,
                      help="Bus or queue we publish resource requests on. [default: %default]")
    parser.add_option("-s", "--notification_subject", dest="notification_subject", type="string",
                      default=DEFAULT_RA_TASK_SPECIFIED_NOTIFICATION_SUBJECT,
                      help="The subject of the event messages which this service publishes. [default: %default]")
    parser.add_option("--otdb_notification_bus", dest="otdb_notification_bus", type="string",
                      default=DEFAULT_OTDB_NOTIFICATION_BUSNAME,
                      help="Bus or queue where the OTDB notifications are published. [default: %default]")
    parser.add_option("--otdb_notification_subject", dest="otdb_notification_subject", type="string",
                      default=DEFAULT_OTDB_NOTIFICATION_SUBJECT,
                      help="Subject of OTDB notifications on otdb_notification_bus. [default: %default]")
    parser.add_option("--otdb_request_bus", dest="otdb_request_bus", type="string",
                      default=DEFAULT_OTDB_SERVICE_BUSNAME,
                      help="Bus or queue where the OTDB requests are handled. [default: %default]")
    parser.add_option('-q', '--broker', dest='broker', type='string', default=None, help='Address of the qpid broker, default: localhost')
    (options, args) = parser.parse_args()

    with RATaskSpecified(otdb_notification_busname=options.otdb_notification_bus,
                         otdb_notification_subject=options.otdb_notification_subject,
                         otdb_service_busname=options.otdb_request_bus,
                         otdb_service_subject=DEFAULT_OTDB_SERVICENAME, ##TODO parse this from command line
                         notification_busname=options.notification_bus,
                         notification_subject=options.notification_subject,
                         broker=options.broker):
        waitForInterrupt()


if __name__ == "__main__":
    main()
