#!/usr/bin/env python

# RABusListener.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id: RABusListener.py 1580 2015-09-30 14:18:57Z loose $

"""
RABusListener listens on the lofar notification message bus and calls (empty) on<SomeMessage> methods when such a message is received.
Typical usage is to derive your own subclass from RABusListener and implement the specific on<SomeMessage> methods that you are interested in.
"""

from lofar.messaging.messagebus import AbstractBusListener
from lofar.sas.resourceassignment.resourceassigner.config import DEFAULT_RA_NOTIFICATION_BUSNAME, DEFAULT_RA_NOTIFICATION_SUBJECTS
from lofar.common.util import waitForInterrupt

import qpid.messaging
import logging
from datetime import datetime

logger = logging.getLogger(__name__)


class RABusListener(AbstractBusListener):
    def __init__(self, busname=DEFAULT_RA_NOTIFICATION_BUSNAME, subjects=DEFAULT_RA_NOTIFICATION_SUBJECTS, broker=None, **kwargs):
        """
        RABusListener listens on the lofar notification message bus and calls (empty) on<SomeMessage> methods when such a message is received.
        Typical usage is to derive your own subclass from RABusListener and implement the specific on<SomeMessage> methods that you are interested in.
        :param busname: valid Qpid address (default: lofar.ra.notification)
        :param broker: valid Qpid broker host (default: None, which means localhost)
        additional parameters in kwargs:
            options=   <dict>  Dictionary of options passed to QPID
            exclusive= <bool>  Create an exclusive binding so no other services can consume duplicate messages (default: False)
            numthreads= <int>  Number of parallel threads processing messages (default: 1)
            verbose=   <bool>  Output extra logging over stdout (default: False)
        """
        self.subject_prefix = (subjects.split('.')[0]+'.') if '.' in subjects else ''

        address = "%s/%s" % (busname, subjects)
        super(RABusListener, self).__init__(address, broker, **kwargs)


    def _handleMessage(self, msg):
        logger.info("on%s: %s" % (msg.subject.replace(self.subject_prefix, ''), str(msg.content).replace('\n', ' ')))

        if msg.subject == '%sTaskScheduled' % self.subject_prefix:
            self.onTaskScheduled(msg.content)
        elif msg.subject == '%sTaskConflict' % self.subject_prefix:
            self.onTaskConflict(msg.content)
        elif msg.subject == '%sTaskError' % self.subject_prefix:
            self.onTaskError(msg.content)
        else:
            logger.error("RABusListener.handleMessage: unknown subject: %s" %str(msg.subject))

    def onTaskScheduled(self, task_ids):
        '''onTaskScheduled is called upon receiving a TaskScheduled message.
        :param task_ids: a dict containing radb_id, mom_id and otdb_id'''
        pass

    def onTaskConflict(self, task_ids):
        '''onTaskConflict is called upon receiving a TaskConflict message.
        :param task_ids: a dict containing radb_id, mom_id and otdb_id'''
        pass

    def onTaskError(self, task_ids):
        '''onTaskError is called upon receiving a TaskError message.
        :param task_ids: a dict containing radb_id, mom_id and otdb_id'''
        pass

if __name__ == '__main__':
    with RABusListener(broker=None) as listener:
        waitForInterrupt()

__all__ = ["RABusListener"]
