#!/usr/bin/python
# $Id$

'''
Class to take a task on approved and add the information needed to put it on prescheduled.
This means adding/updating some Cobal keys, selecting available stations,
selecting the right timeslot and updating start/end time. 
'''
from datetime import datetime, timedelta
import pprint
from lofar.parameterset import PyParameterValue
from lofar.sas.otdb.OTDBBusListener import OTDBBusListener
from lofar.sas.otdb.otdbrpc import OTDBRPC
from lofar.sas.otdb.config import DEFAULT_OTDB_SERVICENAME
from lofar.sas.otdb.config import DEFAULT_OTDB_SERVICE_BUSNAME as DEFAULT_OTDB_BUSNAME
from lofar.sas.otdb.config import DEFAULT_OTDB_NOTIFICATION_BUSNAME, DEFAULT_OTDB_NOTIFICATION_SUBJECT
from lofar.mom.momqueryservice.momqueryrpc import MoMQueryRPC
from lofar.mom.momqueryservice.config import DEFAULT_MOMQUERY_BUSNAME, DEFAULT_MOMQUERY_SERVICENAME
from lofar.sas.resourceassignment.taskprescheduler.cobaltblocksize import CorrelatorSettings, StokesSettings, BlockConstraints, BlockSize
from lofar.sas.resourceassignment.resourceassignmentservice.rpc import RARPC as RADBRPC
from lofar.sas.resourceassignment.resourceassignmentservice.config import DEFAULT_BUSNAME as DEFAULT_RADB_BUSNAME
from lofar.sas.resourceassignment.resourceassignmentservice.config import DEFAULT_SERVICENAME as DEFAULT_RADB_SERVICENAME

import logging
logger = logging.getLogger(__name__)

""" Prefix that is common to all parset keys, when we get a parset from OTDBRPC. """
INPUT_PREFIX = "ObsSW."
""" Prefix that is common to all parset keys, when we need to write a parset to OTDB """
OUTPUT_PREFIX = "LOFAR.ObsSW."
DATAPRODUCTS  = "Observation.DataProducts."
COBALT        = "Observation.ObservationControl.OnlineControl.Cobalt."

def calculateCobaltSettings(parset):
    """uses parset keys to calculate Cobalt block size and integration time"""

    def bool(value):
        return PyParameterValue(str(value), True).getBool()

    if bool(parset[DATAPRODUCTS + "Output_Correlated.enabled"]):
        corr = CorrelatorSettings()
        corr.nrChannelsPerSubband = int(parset[COBALT + "Correlator.nrChannelsPerSubband"])
        corr.integrationTime      = float(parset[COBALT + "Correlator.integrationTime"])
    else:
        corr = None

    if bool(parset[DATAPRODUCTS + "Output_CoherentStokes.enabled"]):
        coherent = StokesSettings()
        coherent.nrChannelsPerSubband  = int(parset[COBALT + "BeamFormer.CoherentStokes.nrChannelsPerSubband"])
        coherent.timeIntegrationFactor = int(parset[COBALT + "BeamFormer.CoherentStokes.timeIntegrationFactor"])
    else:
        coherent = None

    if bool(parset[DATAPRODUCTS + "Output_IncoherentStokes.enabled"]):
        incoherent = StokesSettings()
        incoherent.nrChannelsPerSubband  = int(parset[COBALT + "BeamFormer.IncoherentStokes.nrChannelsPerSubband"])
        incoherent.timeIntegrationFactor = int(parset[COBALT + "BeamFormer.IncoherentStokes.timeIntegrationFactor"])
    else:
        incoherent = None

    clock = int(parset["Observation.sampleClock"])
    constraints = BlockConstraints(corr, coherent, incoherent, clock)
    calculator = BlockSize(constraints)

    return {'nrSubblocks': calculator.nrSubblocks, 'blockSize': calculator.blockSize,
            'nrBlocks': calculator.nrBlocks, 'integrationTime': calculator.integrationTime}

def resourceIndicatorsFromParset( parsetDict ):
    """ Extract the parset keys that are required for determining Cobalt blocksize. """

    #TODO There is some overlap with the RATaskSpecified code here, might need refactoring
    subset = {}

    def get(key, default=None):
        """ Return the value of parset key `key', or `default' if the key
            is not defined. """
        return parsetDict.get(INPUT_PREFIX + key, default)

    def add(key, conversion=lambda x: x):
        """ Add the given key to our subset selection, using an optional
            conversion. """
        value = get(key)
        if value is not None:
            subset[key] = conversion(value)

    """ Some conversion functions for common parameter-value types."""
    #def strvector(value):
    #    return PyParameterValue(str(value), True).getStringVector()

    #def intvector(value):
    #    return PyParameterValue(str(value), True)._expand().getIntVector()

    # =====================================
    # Parset meta info
    # =====================================
    subset["Version.number"] = parsetDict.get("Version.number")

    # =====================================
    # Observation settings
    # =====================================
    add("Observation.momID")
    add("Observation.sampleClock")

    # =====================================
    # Correlator settings
    # =====================================
    add("Observation.DataProducts.Output_Correlated.enabled", bool)
    add("Observation.ObservationControl.OnlineControl.Cobalt.Correlator.integrationTime")
    add("Observation.ObservationControl.OnlineControl.Cobalt.Correlator.nrChannelsPerSubband")

    add("Observation.DataProducts.Output_IncoherentStokes.enabled", bool)
    add("Observation.DataProducts.Output_CoherentStokes.enabled", bool)
    add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.CoherentStokes.nrChannelsPerSubband")
    add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.CoherentStokes.timeIntegrationFactor")
    add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.IncoherentStokes.nrChannelsPerSubband")
    add("Observation.ObservationControl.OnlineControl.Cobalt.BeamFormer.IncoherentStokes.timeIntegrationFactor")

    return subset

def isObservation(parsetDict):
    #TODO copied from RATaskSpecified, needs to be refactored into a common RA lib.
    # Translation table for task types: OTDB (type,subtype) -> RADB (type,subtype)
    OTDB_to_RADB_type_translation_table = {
      ("Observation", "Beam Observation"):    ("observation", "bfmeasurement"),
      ("Observation", "Interferometer"):      ("observation", "interferometer"),
      ("Observation", "TBB (piggyback)"):     ("observation", "tbbmeasurement"),
      ("Observation", "TBB (standalone)"):    ("observation", "tbbmeasurement"),
      ("Pipeline", "Averaging Pipeline"):     ("pipeline", "averaging pipeline"),
      ("Pipeline", "Calibration Pipeline"):   ("pipeline", "calibration pipeline"),
      ("Pipeline", "Imaging Pipeline"):       ("pipeline", "imaging pipeline"),
      ("Pipeline", "Imaging Pipeline MSSS"):  ("pipeline", "imaging pipeline msss"),
      ("Pipeline", "Long Baseline Pipeline"): ("pipeline", "long baseline pipeline"),
      ("Pipeline", "Pulsar Pipeline"):        ("pipeline", "pulsar pipeline"),
      ("MAINTENANCE", ""):                    ("reservation", "maintenance"),
      ("RESERVATION", ""):                    ("reservation", "project"),
    }

    # parse & marshall task type information
    otdb_task_type = parsetDict[INPUT_PREFIX + "Observation.processType"]
    otdb_task_subtype = parsetDict[INPUT_PREFIX + "Observation.processSubtype"]
    task_type, task_subtype = OTDB_to_RADB_type_translation_table[(otdb_task_type, otdb_task_subtype)]
    return (task_type == "observation")

def cobaltOTDBsettings(cobalt_values):
    otdb_info = {}
    otdb_info[OUTPUT_PREFIX + COBALT + "Correlator.nrBlocksPerIntegration"] = cobalt_values["nrBlocks"]
    otdb_info[OUTPUT_PREFIX + COBALT + "Correlator.nrIntegrationsPerBlock"] = cobalt_values["nrSubblocks"]
    otdb_info[OUTPUT_PREFIX + COBALT + "blockSize"] = cobalt_values["blockSize"]
    otdb_info[OUTPUT_PREFIX + COBALT + "Correlator.integrationTime"] = cobalt_values["integrationTime"]
    return otdb_info

class TaskPrescheduler(OTDBBusListener):
    def __init__(self,
                 otdb_notification_busname=DEFAULT_OTDB_NOTIFICATION_BUSNAME,
                 otdb_notification_servicename=DEFAULT_OTDB_NOTIFICATION_SUBJECT,
                 otdb_busname=DEFAULT_OTDB_BUSNAME,
                 otdb_servicename=DEFAULT_OTDB_SERVICENAME,
                 mom_query_busname=DEFAULT_MOMQUERY_BUSNAME,
                 mom_query_servicename=DEFAULT_MOMQUERY_SERVICENAME,
                 radb_busname=DEFAULT_RADB_BUSNAME,
                 radb_servicename=DEFAULT_RADB_SERVICENAME,
                 broker=None, **kwargs):
        super(TaskPrescheduler, self).__init__(busname=otdb_notification_busname,
                                               subject=otdb_notification_servicename,
                                               broker=broker,
                                               **kwargs)

        self.otdbrpc = OTDBRPC(busname=otdb_busname, servicename=otdb_servicename, broker=broker) ## , ForwardExceptions=True hardcoded in RPCWrapper right now
        self.momquery = MoMQueryRPC(busname=mom_query_busname, servicename=mom_query_servicename, timeout=10, broker=broker)
        self.radbrpc = RADBRPC(servicename=radb_servicename, busname=radb_busname, broker=broker, timeout=180)

    def start_listening(self, **kwargs):
        self.otdbrpc.open()
        self.momquery.open()
        self.radbrpc.open()
        super(TaskPrescheduler, self).start_listening(**kwargs)

    def stop_listening(self, **kwargs):
        self.radbrpc.close()
        self.momquery.close()
        self.otdbrpc.close()
        super(TaskPrescheduler, self).stop_listening(**kwargs)

    def onObservationApproved(self, treeId, modificationTime):
        """ Updates task specification and puts task on prescheduled if it was generated by a trigger
        """ #TODO might work for all tasks in the future
        # TODO: only process observations and pipelines
        #TODO we probably need to implement a lot of the checks from
        # std::pair<unscheduled_reasons, QString> Controller::doPreScheduleChecks(Task *task)
        # in Controller.cpp line 2986-3371 from the Alwin Scheduler
        # We might look at the XMLgenerator code as it already does a lot of similar checks.
        # For now we assume the specification is entirely correct.
        # Maybe these checks need to go into the RATaskSpecified instead.
        # NOTE: The MoM predecessor Ids to OTDB predecessor Ids conversion is done in RATaskSpecified on the fly

        otdb_id = treeId
        #Race condition when asking MoM as the mom-otdb-adapter might not have heard that the
        #task is on approved and might still be on approved pending in MoM.
        #mom_ids = self.momquery.getMoMIdsForOTDBIds([otdb_id])
        #So we get the parset for all tasks we receive instead of just for the ones with
        #a trigger.
        try:
            parset = self.otdbrpc.taskGetSpecification( otdb_id=otdb_id )['specification']
            specification = resourceIndicatorsFromParset(parset)
        except Exception as e:
            logger.exception(e)
            logger.error("Problem parsing specification for otdb_id=%s", otdb_id)
            self.radbrpc.updateTaskStatusForOtdbId(otdb_id, 'error') # We don't catch an exception if this fails.
            return
        if specification["Observation.momID"]:
            mom_id = specification["Observation.momID"]
            logger.info('Found mom_id %s for otdb_id %s', mom_id, otdb_id)
        else:
            logger.info('Did not find a mom_id for task otdb_id=%s', otdb_id)
            return
        response = self.momquery.get_trigger_id(mom_id)
        if response['status'] == 'OK':
            logger.info('Found a task mom_id=%s with a trigger_id=%s', mom_id, response['trigger_id'])
            
            #TODO, check for stations and other resources, start/endtime, target position, then update specification

            logger.info('prescheduling otdb_id=%s because it was generated by trigger_id=%s', otdb_id, response['trigger_id'])
            otdb_info = {}
            if isObservation(parset):
                cobalt_values = calculateCobaltSettings(specification)
                otdb_info.update(cobaltOTDBsettings(cobalt_values))
            self.setOTDBinfo(otdb_id, otdb_info, 'prescheduled')
        else:
            logger.info('Did not find a trigger for task mom_id=%s, because %s', mom_id, response['errors'])
            


    def setOTDBinfo(self, otdb_id, otdb_info, otdb_status):
        """This function sets the values in otdb_info in OTDB, almost a copy from the RAtoOTDBPropagator"""
        try:
            if otdb_info:
                logger.info('Setting specification for otdb_id %s:\n', otdb_id)
                logger.info(pprint.pformat(otdb_info))
                self.otdbrpc.taskSetSpecification(otdb_id, otdb_info)
                #We probably will need this as well
                #self.otdbrpc.taskPrepareForScheduling(otdb_id, otdb_info["LOFAR.ObsSW.Observation.startTime"],
                #                                               otdb_info["LOFAR.ObsSW.Observation.stopTime"])
            logger.info('Setting status (%s) for otdb_id %s', otdb_status, otdb_id)
            self.otdbrpc.taskSetStatus(otdb_id, otdb_status)
        except Exception as e:
            logger.exception(e)
            logger.error("Problem setting specification or status in OTDB for otdb_id=%s", otdb_id)
            self.radbrpc.updateTaskStatusForOtdbId(otdb_id, 'error') # We don't catch an exception if this fails.

def main():
    from optparse import OptionParser
    from lofar.messaging import setQpidLogLevel
    from lofar.common.util import waitForInterrupt

    # make sure we run in UTC timezone
    import os
    os.environ['TZ'] = 'UTC'

    # Check the invocation arguments
    parser = OptionParser("%prog [options]", description='runs the task prescheduler service')
    parser.add_option('-q', '--broker', dest='broker', type='string', default=None, help='Address of the qpid broker, default: localhost')
    parser.add_option('-V', '--verbose', dest='verbose', action='store_true', help='verbose logging')
    parser.add_option("--otdb_notification_busname", dest="otdb_notification_busname", type="string",
                      default=DEFAULT_OTDB_NOTIFICATION_BUSNAME,
                      help="Bus or queue where the OTDB notifications are published. [default: %default]")
    parser.add_option("--otdb_notification_servicename", dest="otdb_notification_servicename", type="string",
                      default=DEFAULT_OTDB_NOTIFICATION_SUBJECT,
                      help="Subject of OTDB notifications on otdb_notification_busname. [default: %default]")
    parser.add_option("--otdb_busname", dest="otdb_busname", type="string",
                      default=DEFAULT_OTDB_BUSNAME,
                      help="Name of the OTDB bus exchange on the qpid broker. [default: %default]")
    parser.add_option("--otdb_servicename", dest="otdb_servicename", type="string",
                      default=DEFAULT_OTDB_SERVICENAME,
                      help="Name of the OTDB service. [default: %default]")
    parser.add_option("--momquery_busname", dest="momquery_busname", type="string",
                      default=DEFAULT_MOMQUERY_BUSNAME,
                      help="Name of the momquery bus exchange on the qpid broker. [default: %default]")
    parser.add_option("--momquery_servicename", dest="momquery_servicename", type="string",
                      default=DEFAULT_MOMQUERY_SERVICENAME,
                      help="Name of the momquery service. [default: %default]")
    parser.add_option("--radb_busname", dest="radb_busname", type="string",
                      default=DEFAULT_RADB_BUSNAME,
                      help="Name of the RADB bus exchange on the qpid broker. [default: %default]")
    parser.add_option("--radb_servicename", dest="radb_servicename", type="string",
                      default=DEFAULT_RADB_SERVICENAME,
                      help="Name of the RADB service. [default: %default]")
    (options, args) = parser.parse_args()

    setQpidLogLevel(logging.INFO)
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.DEBUG if options.verbose else logging.INFO)

    with TaskPrescheduler(otdb_notification_busname=options.otdb_notification_busname,
                                      otdb_notification_servicename=options.otdb_notification_servicename,
                                      otdb_busname=options.otdb_busname,
                                      otdb_servicename=options.otdb_servicename,
                                      mom_query_busname=options.momquery_busname,
                                      mom_query_servicename=options.momquery_servicename,
                                      radb_busname=options.radb_busname,
                                      radb_servicename=options.radb_servicename,
                                      broker=options.broker):
        waitForInterrupt()

if __name__ == '__main__':
    main()
