#!/usr/bin/env python

import unittest
import uuid
import datetime
from lofar.messaging import Service
from lofar.sas.systemstatus.service.SSDBQueryService import DataMonitorQueryService, createService
from lofar.sas.systemstatus.service.SSDBrpc import SSDBRPC

try:
    from qpid.messaging import Connection
    from qpidtoollibs import BrokerAgent
except ImportError:
    print 'Cannot run test without qpid tools'
    print 'Please source qpid profile'
    exit(3)

try:
    from mock import MagicMock
    from mock import patch
except ImportError:
    print 'Cannot run test without python MagicMock'
    print 'Please install MagicMock: pip install mock'
    exit(3)

connection = None
broker = None

try:
    # setup broker connection
    connection = Connection.establish('127.0.0.1')
    broker = BrokerAgent(connection)

    # add test service busname
    busname = 'test-lofarbus-%s' % (uuid.uuid1())
    broker.addExchange('topic', busname)
    servicename = "test-SSDB-service"

    # the system under test is the service and the rpc, not the SSDatabase
    # so, patch (mock) the SSDatabase class during these tests.
    # when the service instantiates an SSDatabase it will get the mocked class.
    with patch('lofar.sas.systemstatus.service.SSDBQueryService.SSDB', autospec=True) as MockSSDB:
        mock = MockSSDB.return_value
        # modify the return values of the various SSDatabase methods with pre-cooked answers
        def ensure_connected(self):
            self.DBconnected = (self.conn and self.conn.status==1)
            if not self.DBconnected:
                try:
                    self.conn= pg.connect("dbname=%s user=%s password=%s" % (DATABASE,USER,PASSWORD))
                    self.DBconnected = (self.conn and self.conn.status==1)
                except Exception as e:
                    logger.error("DB connection could not be restored.")
            return self.DBconnected
        mock.ensure_connected.returnvalue=True

        mock.getstatenames.return_value=[{'statename': 'Inactive', 'id': 0L}, {'statename': 'Active', 'id': 1L}]
        mock.getactivegroupnames.return_value=[{'groupname': 'storagenodes', 'id': 0L}, {'groupname': 'computenodes', 'id': 1L},
                {'groupname': 'archivenodes', 'id': 2L}, {'groupname': 'locusnodes', 'id': 3L}, {'groupname': 'cep4', 'id': 4L}]
        mock.gethostsforgid.return_value=[{'statename': 'Active', 'totalspace': 702716L, 'hostname': 'lustre001', 'usedspace': 23084L,
                                                       'groupname': 'cep4', 'claimedspace': 0L, 'path': '/lustre', 'id': 1L}]
        mock.gethostsforgroups.return_value=[{'hostname': 'lse001', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse002',
                                               'groupid': 0L, 'statusid': 0L}, {'hostname': 'lse003', 'groupid': 0L, 'statusid': 1L},
                                               {'hostname': 'lse004', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse005', 'groupid': 0L,
                                               'statusid': 1L}, {'hostname': 'lse006', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse007',
                                               'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse008', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse009', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse010', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse011', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse012', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse013', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse014', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse015', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse016', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse017', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse018', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse019', 'groupid': 0L, 'statusid': 0L}, {'hostname': 'lse020', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse021', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse022', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse023', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lse024', 'groupid': 0L, 'statusid': 1L}, {'hostname': 'lce001', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce002', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce003', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce004', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce005', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce006', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce007', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce008', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce009', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce010', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce011', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce012', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce013', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce014', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce015', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce016', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce017', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce018', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce019', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce020', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce021', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce022', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce023', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce024', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce025', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce026', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce027', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce028', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce029', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce030', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce031', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce032', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce033', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce034', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce035', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce036', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce037', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce038', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce039', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce040', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce041', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce042', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce043', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce044', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce045', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce046', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce047', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce048', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce049', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce050', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce051', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce052', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce053', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce054', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce055', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce056', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce057', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce058', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce059', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce060', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce061', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce062', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce063', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce064', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce065', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce066', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce067', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce068', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce069', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce070', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce071', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lce072', 'groupid': 1L, 'statusid': 0L}, {'hostname': 'lexar001', 'groupid': 2L, 'statusid': 0L}, {'hostname': 'lexar002', 'groupid': 2L, 'statusid': 0L}, {'hostname': 'locus001', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus002', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus003', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus004', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus005', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus006', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus007', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus008', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus009', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus010', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus011', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus012', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus013', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus014', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus015', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus016', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus017', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus018', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus019', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus020', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus021', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus022', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus023', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus024', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus025', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus026', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus027', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus028', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus029', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus030', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus031', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus032', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus033', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus034', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus035', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus036', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus037', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus038', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus039', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus040', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus041', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus042', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus043', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus044', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus045', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus046', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus047', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus048', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus049', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus050', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus051', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus052', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus053', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus054', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus055', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus056', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus057', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus058', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus059', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus060', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus061', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus062', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus063', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus064', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus065', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus066', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus067', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus068', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus069', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus070', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus071', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus072', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus073', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus074', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus075', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus076', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus077', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus078', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus079', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus080', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus081', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus082', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus083', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus084', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus085', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus086', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus087', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus088', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus089', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus090', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus091', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus092', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus093', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus094', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus095', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus096', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus097', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus098', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus099', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'locus100', 'groupid': 3L, 'statusid': 0L}, {'hostname': 'lustre001', 'groupid': 4L, 'statusid': 1L}]

        mock.listall.return_value=[{'hostid': 1L, 'totalspace': 0L, 'hostname': 'lse001', 'usedspace': 0L, 'id': 1L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 1L, 'totalspace': 0L, 'hostname': 'lse001', 'usedspace': 0L, 'id': 2L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 1L, 'totalspace': 0L, 'hostname': 'lse001', 'usedspace': 0L, 'id': 3L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 1L, 'totalspace': 0L, 'hostname': 'lse001', 'usedspace': 0L, 'id': 4L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 2L, 'totalspace': 0L, 'hostname': 'lse002', 'usedspace': 0L, 'id': 5L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 0L}, {'hostid': 2L, 'totalspace': 0L, 'hostname': 'lse002', 'usedspace': 0L, 'id': 6L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 0L}, {'hostid': 2L, 'totalspace': 0L, 'hostname': 'lse002', 'usedspace': 0L, 'id': 7L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 0L}, {'hostid': 2L, 'totalspace': 0L, 'hostname': 'lse002', 'usedspace': 0L, 'id': 8L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 0L}, {'hostid': 3L, 'totalspace': 0L, 'hostname': 'lse003', 'usedspace': 0L, 'id': 9L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 3L, 'totalspace': 0L, 'hostname': 'lse003', 'usedspace': 0L, 'id': 10L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 3L, 'totalspace': 0L, 'hostname': 'lse003', 'usedspace': 0L, 'id': 11L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 3L, 'totalspace': 0L, 'hostname': 'lse003', 'usedspace': 0L, 'id': 12L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 4L, 'totalspace': 0L, 'hostname': 'lse004', 'usedspace': 0L, 'id': 13L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 4L, 'totalspace': 0L, 'hostname': 'lse004', 'usedspace': 0L, 'id': 14L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 4L, 'totalspace': 0L, 'hostname': 'lse004', 'usedspace': 0L, 'id': 15L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 4L, 'totalspace': 0L, 'hostname': 'lse004', 'usedspace': 0L, 'id': 16L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 5L, 'totalspace': 0L, 'hostname': 'lse005', 'usedspace': 0L, 'id': 17L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 5L, 'totalspace': 0L, 'hostname': 'lse005', 'usedspace': 0L, 'id': 18L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 5L, 'totalspace': 0L, 'hostname': 'lse005', 'usedspace': 0L, 'id': 19L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 5L, 'totalspace': 0L, 'hostname': 'lse005', 'usedspace': 0L, 'id': 20L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 6L, 'totalspace': 0L, 'hostname': 'lse006', 'usedspace': 0L, 'id': 21L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 6L, 'totalspace': 0L, 'hostname': 'lse006', 'usedspace': 0L, 'id': 22L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 6L, 'totalspace': 0L, 'hostname': 'lse006', 'usedspace': 0L, 'id': 23L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 6L, 'totalspace': 0L, 'hostname': 'lse006', 'usedspace': 0L, 'id': 24L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 7L, 'totalspace': 0L, 'hostname': 'lse007', 'usedspace': 0L, 'id': 25L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 7L, 'totalspace': 0L, 'hostname': 'lse007', 'usedspace': 0L, 'id': 26L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 7L, 'totalspace': 0L, 'hostname': 'lse007', 'usedspace': 0L, 'id': 27L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 7L, 'totalspace': 0L, 'hostname': 'lse007', 'usedspace': 0L, 'id': 28L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 8L, 'totalspace': 0L, 'hostname': 'lse008', 'usedspace': 0L, 'id': 29L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 8L, 'totalspace': 0L, 'hostname': 'lse008', 'usedspace': 0L, 'id': 30L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 8L, 'totalspace': 0L, 'hostname': 'lse008', 'usedspace': 0L, 'id': 31L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 8L, 'totalspace': 0L, 'hostname': 'lse008', 'usedspace': 0L, 'id': 32L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 9L, 'totalspace': 0L, 'hostname': 'lse009', 'usedspace': 0L, 'id': 33L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 9L, 'totalspace': 0L, 'hostname': 'lse009', 'usedspace': 0L, 'id': 34L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 9L, 'totalspace': 0L, 'hostname': 'lse009', 'usedspace': 0L, 'id': 35L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 9L, 'totalspace': 0L, 'hostname': 'lse009', 'usedspace': 0L, 'id': 36L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 10L, 'totalspace': 0L, 'hostname': 'lse010', 'usedspace': 0L, 'id': 37L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 10L, 'totalspace': 0L, 'hostname': 'lse010', 'usedspace': 0L, 'id': 38L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 10L, 'totalspace': 0L, 'hostname': 'lse010', 'usedspace': 0L, 'id': 39L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 10L, 'totalspace': 0L, 'hostname': 'lse010', 'usedspace': 0L, 'id': 40L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 11L, 'totalspace': 0L, 'hostname': 'lse011', 'usedspace': 0L, 'id': 41L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 11L, 'totalspace': 0L, 'hostname': 'lse011', 'usedspace': 0L, 'id': 42L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 11L, 'totalspace': 0L, 'hostname': 'lse011', 'usedspace': 0L, 'id': 43L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 11L, 'totalspace': 0L, 'hostname': 'lse011', 'usedspace': 0L, 'id': 44L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 12L, 'totalspace': 0L, 'hostname': 'lse012', 'usedspace': 0L, 'id': 45L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 12L, 'totalspace': 0L, 'hostname': 'lse012', 'usedspace': 0L, 'id': 46L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 12L, 'totalspace': 0L, 'hostname': 'lse012', 'usedspace': 0L, 'id': 47L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 12L, 'totalspace': 1L, 'hostname': 'lse012', 'usedspace': 1L, 'id': 48L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 13L, 'totalspace': 0L, 'hostname': 'lse013', 'usedspace': 0L, 'id': 49L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 13L, 'totalspace': 0L, 'hostname': 'lse013', 'usedspace': 0L, 'id': 50L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 13L, 'totalspace': 0L, 'hostname': 'lse013', 'usedspace': 0L, 'id': 51L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 13L, 'totalspace': 0L, 'hostname': 'lse013', 'usedspace': 0L, 'id': 52L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 14L, 'totalspace': 0L, 'hostname': 'lse014', 'usedspace': 0L, 'id': 53L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 14L, 'totalspace': 0L, 'hostname': 'lse014', 'usedspace': 0L, 'id': 54L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 14L, 'totalspace': 0L, 'hostname': 'lse014', 'usedspace': 0L, 'id': 55L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 14L, 'totalspace': 1L, 'hostname': 'lse014', 'usedspace': 1L, 'id': 56L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 15L, 'totalspace': 0L, 'hostname': 'lse015', 'usedspace': 0L, 'id': 57L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 15L, 'totalspace': 0L, 'hostname': 'lse015', 'usedspace': 0L, 'id': 58L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 15L, 'totalspace': 0L, 'hostname': 'lse015', 'usedspace': 0L, 'id': 59L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 15L, 'totalspace': 0L, 'hostname': 'lse015', 'usedspace': 0L, 'id': 60L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 16L, 'totalspace': 0L, 'hostname': 'lse016', 'usedspace': 0L, 'id': 61L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 16L, 'totalspace': 0L, 'hostname': 'lse016', 'usedspace': 0L, 'id': 62L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 16L, 'totalspace': 0L, 'hostname': 'lse016', 'usedspace': 0L, 'id': 63L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 16L, 'totalspace': 1L, 'hostname': 'lse016', 'usedspace': 1L, 'id': 64L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 17L, 'totalspace': 0L, 'hostname': 'lse017', 'usedspace': 0L, 'id': 65L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 17L, 'totalspace': 0L, 'hostname': 'lse017', 'usedspace': 0L, 'id': 66L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 17L, 'totalspace': 0L, 'hostname': 'lse017', 'usedspace': 0L, 'id': 67L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 17L, 'totalspace': 1L, 'hostname': 'lse017', 'usedspace': 1L, 'id': 68L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 18L, 'totalspace': 0L, 'hostname': 'lse018', 'usedspace': 0L, 'id': 69L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 18L, 'totalspace': 0L, 'hostname': 'lse018', 'usedspace': 0L, 'id': 70L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 18L, 'totalspace': 0L, 'hostname': 'lse018', 'usedspace': 0L, 'id': 71L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 18L, 'totalspace': 1L, 'hostname': 'lse018', 'usedspace': 1L, 'id': 72L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 19L, 'totalspace': 0L, 'hostname': 'lse019', 'usedspace': 0L, 'id': 73L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 0L}, {'hostid': 19L, 'totalspace': 0L, 'hostname': 'lse019', 'usedspace': 0L, 'id': 74L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 0L}, {'hostid': 19L, 'totalspace': 0L, 'hostname': 'lse019', 'usedspace': 0L, 'id': 75L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 0L}, {'hostid': 19L, 'totalspace': 1L, 'hostname': 'lse019', 'usedspace': 1L, 'id': 76L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 0L}, {'hostid': 20L, 'totalspace': 0L, 'hostname': 'lse020', 'usedspace': 0L, 'id': 77L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 20L, 'totalspace': 0L, 'hostname': 'lse020', 'usedspace': 0L, 'id': 78L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 20L, 'totalspace': 0L, 'hostname': 'lse020', 'usedspace': 0L, 'id': 79L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 20L, 'totalspace': 1L, 'hostname': 'lse020', 'usedspace': 1L, 'id': 80L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 21L, 'totalspace': 0L, 'hostname': 'lse021', 'usedspace': 0L, 'id': 81L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 21L, 'totalspace': 0L, 'hostname': 'lse021', 'usedspace': 0L, 'id': 82L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 21L, 'totalspace': 0L, 'hostname': 'lse021', 'usedspace': 0L, 'id': 83L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 21L, 'totalspace': 1L, 'hostname': 'lse021', 'usedspace': 1L, 'id': 84L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 22L, 'totalspace': 0L, 'hostname': 'lse022', 'usedspace': 0L, 'id': 85L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 22L, 'totalspace': 0L, 'hostname': 'lse022', 'usedspace': 0L, 'id': 86L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 22L, 'totalspace': 0L, 'hostname': 'lse022', 'usedspace': 0L, 'id': 87L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 22L, 'totalspace': 1L, 'hostname': 'lse022', 'usedspace': 1L, 'id': 88L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 23L, 'totalspace': 0L, 'hostname': 'lse023', 'usedspace': 0L, 'id': 89L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 23L, 'totalspace': 0L, 'hostname': 'lse023', 'usedspace': 0L, 'id': 90L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 23L, 'totalspace': 0L, 'hostname': 'lse023', 'usedspace': 0L, 'id': 91L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 23L, 'totalspace': 1L, 'hostname': 'lse023', 'usedspace': 1L, 'id': 92L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 24L, 'totalspace': 0L, 'hostname': 'lse024', 'usedspace': 0L, 'id': 93L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data1', 'statusid': 1L}, {'hostid': 24L, 'totalspace': 0L, 'hostname': 'lse024', 'usedspace': 0L, 'id': 94L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data2', 'statusid': 1L}, {'hostid': 24L, 'totalspace': 0L, 'hostname': 'lse024', 'usedspace': 0L, 'id': 95L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data3', 'statusid': 1L}, {'hostid': 24L, 'totalspace': 1L, 'hostname': 'lse024', 'usedspace': 1L, 'id': 96L, 'groupid': 0L, 'claimedspace': 0L, 'path': '/data4', 'statusid': 1L}, {'hostid': 25L, 'totalspace': 0L, 'hostname': 'lce001', 'usedspace': 0L, 'id': 97L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 26L, 'totalspace': 0L, 'hostname': 'lce002', 'usedspace': 0L, 'id': 98L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 27L, 'totalspace': 0L, 'hostname': 'lce003', 'usedspace': 0L, 'id': 99L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 28L, 'totalspace': 0L, 'hostname': 'lce004', 'usedspace': 0L, 'id': 100L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 29L, 'totalspace': 0L, 'hostname': 'lce005', 'usedspace': 0L, 'id': 101L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 30L, 'totalspace': 0L, 'hostname': 'lce006', 'usedspace': 0L, 'id': 102L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 31L, 'totalspace': 0L, 'hostname': 'lce007', 'usedspace': 0L, 'id': 103L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 32L, 'totalspace': 1L, 'hostname': 'lce008', 'usedspace': 1L, 'id': 104L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 33L, 'totalspace': 1L, 'hostname': 'lce009', 'usedspace': 1L, 'id': 105L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 34L, 'totalspace': 1L, 'hostname': 'lce010', 'usedspace': 1L, 'id': 106L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 35L, 'totalspace': 1L, 'hostname': 'lce011', 'usedspace': 1L, 'id': 107L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 36L, 'totalspace': 1L, 'hostname': 'lce012', 'usedspace': 1L, 'id': 108L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 37L, 'totalspace': 1L, 'hostname': 'lce013', 'usedspace': 1L, 'id': 109L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 38L, 'totalspace': 1L, 'hostname': 'lce014', 'usedspace': 1L, 'id': 110L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 39L, 'totalspace': 1L, 'hostname': 'lce015', 'usedspace': 1L, 'id': 111L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 40L, 'totalspace': 1L, 'hostname': 'lce016', 'usedspace': 1L, 'id': 112L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 41L, 'totalspace': 1L, 'hostname': 'lce017', 'usedspace': 1L, 'id': 113L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 42L, 'totalspace': 1L, 'hostname': 'lce018', 'usedspace': 1L, 'id': 114L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 43L, 'totalspace': 1L, 'hostname': 'lce019', 'usedspace': 1L, 'id': 115L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 44L, 'totalspace': 1L, 'hostname': 'lce020', 'usedspace': 1L, 'id': 116L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 45L, 'totalspace': 1L, 'hostname': 'lce021', 'usedspace': 0L, 'id': 117L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 46L, 'totalspace': 1L, 'hostname': 'lce022', 'usedspace': 1L, 'id': 118L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 47L, 'totalspace': 1L, 'hostname': 'lce023', 'usedspace': 1L, 'id': 119L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 48L, 'totalspace': 1L, 'hostname': 'lce024', 'usedspace': 1L, 'id': 120L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 49L, 'totalspace': 1L, 'hostname': 'lce025', 'usedspace': 1L, 'id': 121L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 50L, 'totalspace': 1L, 'hostname': 'lce026', 'usedspace': 1L, 'id': 122L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 51L, 'totalspace': 1L, 'hostname': 'lce027', 'usedspace': 1L, 'id': 123L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 52L, 'totalspace': 1L, 'hostname': 'lce028', 'usedspace': 1L, 'id': 124L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 53L, 'totalspace': 1L, 'hostname': 'lce029', 'usedspace': 1L, 'id': 125L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 54L, 'totalspace': 1L, 'hostname': 'lce030', 'usedspace': 1L, 'id': 126L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 55L, 'totalspace': 1L, 'hostname': 'lce031', 'usedspace': 1L, 'id': 127L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 56L, 'totalspace': 1L, 'hostname': 'lce032', 'usedspace': 1L, 'id': 128L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 57L, 'totalspace': 1L, 'hostname': 'lce033', 'usedspace': 1L, 'id': 129L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 58L, 'totalspace': 1L, 'hostname': 'lce034', 'usedspace': 1L, 'id': 130L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 59L, 'totalspace': 1L, 'hostname': 'lce035', 'usedspace': 1L, 'id': 131L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 60L, 'totalspace': 1L, 'hostname': 'lce036', 'usedspace': 1L, 'id': 132L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 61L, 'totalspace': 1L, 'hostname': 'lce037', 'usedspace': 1L, 'id': 133L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 62L, 'totalspace': 1L, 'hostname': 'lce038', 'usedspace': 1L, 'id': 134L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 63L, 'totalspace': 1L, 'hostname': 'lce039', 'usedspace': 1L, 'id': 135L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 64L, 'totalspace': 1L, 'hostname': 'lce040', 'usedspace': 1L, 'id': 136L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 65L, 'totalspace': 1L, 'hostname': 'lce041', 'usedspace': 1L, 'id': 137L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 66L, 'totalspace': 1L, 'hostname': 'lce042', 'usedspace': 1L, 'id': 138L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 67L, 'totalspace': 1L, 'hostname': 'lce043', 'usedspace': 1L, 'id': 139L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 68L, 'totalspace': 1L, 'hostname': 'lce044', 'usedspace': 1L, 'id': 140L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 69L, 'totalspace': 1L, 'hostname': 'lce045', 'usedspace': 1L, 'id': 141L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 70L, 'totalspace': 1L, 'hostname': 'lce046', 'usedspace': 1L, 'id': 142L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 71L, 'totalspace': 1L, 'hostname': 'lce047', 'usedspace': 1L, 'id': 143L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 72L, 'totalspace': 1L, 'hostname': 'lce048', 'usedspace': 1L, 'id': 144L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 73L, 'totalspace': 1L, 'hostname': 'lce049', 'usedspace': 1L, 'id': 145L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 74L, 'totalspace': 1L, 'hostname': 'lce050', 'usedspace': 1L, 'id': 146L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 75L, 'totalspace': 1L, 'hostname': 'lce051', 'usedspace': 1L, 'id': 147L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 76L, 'totalspace': 1L, 'hostname': 'lce052', 'usedspace': 1L, 'id': 148L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 77L, 'totalspace': 1L, 'hostname': 'lce053', 'usedspace': 1L, 'id': 149L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 78L, 'totalspace': 1L, 'hostname': 'lce054', 'usedspace': 1L, 'id': 150L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 79L, 'totalspace': 1L, 'hostname': 'lce055', 'usedspace': 1L, 'id': 151L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 80L, 'totalspace': 1L, 'hostname': 'lce056', 'usedspace': 1L, 'id': 152L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 81L, 'totalspace': 1L, 'hostname': 'lce057', 'usedspace': 1L, 'id': 153L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 82L, 'totalspace': 1L, 'hostname': 'lce058', 'usedspace': 1L, 'id': 154L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 83L, 'totalspace': 1L, 'hostname': 'lce059', 'usedspace': 1L, 'id': 155L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 84L, 'totalspace': 1L, 'hostname': 'lce060', 'usedspace': 1L, 'id': 156L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 85L, 'totalspace': 1L, 'hostname': 'lce061', 'usedspace': 1L, 'id': 157L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 86L, 'totalspace': 1L, 'hostname': 'lce062', 'usedspace': 1L, 'id': 158L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 87L, 'totalspace': 1L, 'hostname': 'lce063', 'usedspace': 1L, 'id': 159L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 88L, 'totalspace': 1L, 'hostname': 'lce064', 'usedspace': 1L, 'id': 160L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 89L, 'totalspace': 1L, 'hostname': 'lce065', 'usedspace': 1L, 'id': 161L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 90L, 'totalspace': 1L, 'hostname': 'lce066', 'usedspace': 1L, 'id': 162L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 91L, 'totalspace': 1L, 'hostname': 'lce067', 'usedspace': 1L, 'id': 163L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 92L, 'totalspace': 1L, 'hostname': 'lce068', 'usedspace': 1L, 'id': 164L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 93L, 'totalspace': 1L, 'hostname': 'lce069', 'usedspace': 1L, 'id': 165L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 94L, 'totalspace': 1L, 'hostname': 'lce070', 'usedspace': 1L, 'id': 166L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 95L, 'totalspace': 1L, 'hostname': 'lce071', 'usedspace': 1L, 'id': 167L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 96L, 'totalspace': 1L, 'hostname': 'lce072', 'usedspace': 1L, 'id': 168L, 'groupid': 1L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 97L, 'totalspace': 1L, 'hostname': 'lexar001', 'usedspace': 1L, 'id': 169L, 'groupid': 2L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 98L, 'totalspace': 1L, 'hostname': 'lexar002', 'usedspace': 1L, 'id': 170L, 'groupid': 2L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 99L, 'totalspace': 1L, 'hostname': 'locus001', 'usedspace': 1L, 'id': 171L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 100L, 'totalspace': 1L, 'hostname': 'locus002', 'usedspace': 1L, 'id': 172L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 101L, 'totalspace': 1L, 'hostname': 'locus003', 'usedspace': 1L, 'id': 173L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 102L, 'totalspace': 1L, 'hostname': 'locus004', 'usedspace': 1L, 'id': 174L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 103L, 'totalspace': 1L, 'hostname': 'locus005', 'usedspace': 1L, 'id': 175L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 104L, 'totalspace': 1L, 'hostname': 'locus006', 'usedspace': 1L, 'id': 176L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 105L, 'totalspace': 1L, 'hostname': 'locus007', 'usedspace': 1L, 'id': 177L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 106L, 'totalspace': 1L, 'hostname': 'locus008', 'usedspace': 1L, 'id': 178L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 107L, 'totalspace': 1L, 'hostname': 'locus009', 'usedspace': 1L, 'id': 179L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 108L, 'totalspace': 1L, 'hostname': 'locus010', 'usedspace': 1L, 'id': 180L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 109L, 'totalspace': 1L, 'hostname': 'locus011', 'usedspace': 1L, 'id': 181L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 110L, 'totalspace': 1L, 'hostname': 'locus012', 'usedspace': 1L, 'id': 182L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 111L, 'totalspace': 1L, 'hostname': 'locus013', 'usedspace': 1L, 'id': 183L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 112L, 'totalspace': 1L, 'hostname': 'locus014', 'usedspace': 1L, 'id': 184L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 113L, 'totalspace': 1L, 'hostname': 'locus015', 'usedspace': 1L, 'id': 185L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 114L, 'totalspace': 1L, 'hostname': 'locus016', 'usedspace': 1L, 'id': 186L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 115L, 'totalspace': 1L, 'hostname': 'locus017', 'usedspace': 1L, 'id': 187L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 116L, 'totalspace': 1L, 'hostname': 'locus018', 'usedspace': 1L, 'id': 188L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 117L, 'totalspace': 1L, 'hostname': 'locus019', 'usedspace': 1L, 'id': 189L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 118L, 'totalspace': 1L, 'hostname': 'locus020', 'usedspace': 1L, 'id': 190L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 119L, 'totalspace': 1L, 'hostname': 'locus021', 'usedspace': 1L, 'id': 191L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 120L, 'totalspace': 1L, 'hostname': 'locus022', 'usedspace': 1L, 'id': 192L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 121L, 'totalspace': 1L, 'hostname': 'locus023', 'usedspace': 1L, 'id': 193L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 122L, 'totalspace': 1L, 'hostname': 'locus024', 'usedspace': 1L, 'id': 194L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 123L, 'totalspace': 1L, 'hostname': 'locus025', 'usedspace': 1L, 'id': 195L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 124L, 'totalspace': 1L, 'hostname': 'locus026', 'usedspace': 1L, 'id': 196L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 125L, 'totalspace': 1L, 'hostname': 'locus027', 'usedspace': 1L, 'id': 197L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 126L, 'totalspace': 1L, 'hostname': 'locus028', 'usedspace': 1L, 'id': 198L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 127L, 'totalspace': 1L, 'hostname': 'locus029', 'usedspace': 1L, 'id': 199L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 128L, 'totalspace': 1L, 'hostname': 'locus030', 'usedspace': 1L, 'id': 200L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 129L, 'totalspace': 1L, 'hostname': 'locus031', 'usedspace': 1L, 'id': 201L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 130L, 'totalspace': 1L, 'hostname': 'locus032', 'usedspace': 1L, 'id': 202L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 131L, 'totalspace': 1L, 'hostname': 'locus033', 'usedspace': 1L, 'id': 203L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 132L, 'totalspace': 1L, 'hostname': 'locus034', 'usedspace': 1L, 'id': 204L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 133L, 'totalspace': 1L, 'hostname': 'locus035', 'usedspace': 1L, 'id': 205L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 134L, 'totalspace': 1L, 'hostname': 'locus036', 'usedspace': 1L, 'id': 206L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 135L, 'totalspace': 1L, 'hostname': 'locus037', 'usedspace': 1L, 'id': 207L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 136L, 'totalspace': 1L, 'hostname': 'locus038', 'usedspace': 1L, 'id': 208L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 137L, 'totalspace': 1L, 'hostname': 'locus039', 'usedspace': 1L, 'id': 209L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 138L, 'totalspace': 1L, 'hostname': 'locus040', 'usedspace': 1L, 'id': 210L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 139L, 'totalspace': 1L, 'hostname': 'locus041', 'usedspace': 1L, 'id': 211L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 140L, 'totalspace': 1L, 'hostname': 'locus042', 'usedspace': 1L, 'id': 212L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 141L, 'totalspace': 1L, 'hostname': 'locus043', 'usedspace': 1L, 'id': 213L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 142L, 'totalspace': 1L, 'hostname': 'locus044', 'usedspace': 1L, 'id': 214L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 143L, 'totalspace': 1L, 'hostname': 'locus045', 'usedspace': 1L, 'id': 215L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 144L, 'totalspace': 1L, 'hostname': 'locus046', 'usedspace': 1L, 'id': 216L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 145L, 'totalspace': 1L, 'hostname': 'locus047', 'usedspace': 1L, 'id': 217L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 146L, 'totalspace': 1L, 'hostname': 'locus048', 'usedspace': 1L, 'id': 218L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 147L, 'totalspace': 1L, 'hostname': 'locus049', 'usedspace': 1L, 'id': 219L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 148L, 'totalspace': 1L, 'hostname': 'locus050', 'usedspace': 1L, 'id': 220L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 149L, 'totalspace': 1L, 'hostname': 'locus051', 'usedspace': 1L, 'id': 221L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 150L, 'totalspace': 1L, 'hostname': 'locus052', 'usedspace': 1L, 'id': 222L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 151L, 'totalspace': 1L, 'hostname': 'locus053', 'usedspace': 1L, 'id': 223L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 152L, 'totalspace': 1L, 'hostname': 'locus054', 'usedspace': 1L, 'id': 224L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 153L, 'totalspace': 1L, 'hostname': 'locus055', 'usedspace': 1L, 'id': 225L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 154L, 'totalspace': 1L, 'hostname': 'locus056', 'usedspace': 1L, 'id': 226L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 155L, 'totalspace': 1L, 'hostname': 'locus057', 'usedspace': 1L, 'id': 227L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 156L, 'totalspace': 1L, 'hostname': 'locus058', 'usedspace': 1L, 'id': 228L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 157L, 'totalspace': 1L, 'hostname': 'locus059', 'usedspace': 1L, 'id': 229L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 158L, 'totalspace': 1L, 'hostname': 'locus060', 'usedspace': 1L, 'id': 230L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 159L, 'totalspace': 1L, 'hostname': 'locus061', 'usedspace': 1L, 'id': 231L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 160L, 'totalspace': 1L, 'hostname': 'locus062', 'usedspace': 1L, 'id': 232L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 161L, 'totalspace': 1L, 'hostname': 'locus063', 'usedspace': 1L, 'id': 233L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 162L, 'totalspace': 1L, 'hostname': 'locus064', 'usedspace': 1L, 'id': 234L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 163L, 'totalspace': 1L, 'hostname': 'locus065', 'usedspace': 1L, 'id': 235L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 164L, 'totalspace': 1L, 'hostname': 'locus066', 'usedspace': 1L, 'id': 236L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 165L, 'totalspace': 1L, 'hostname': 'locus067', 'usedspace': 1L, 'id': 237L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 166L, 'totalspace': 1L, 'hostname': 'locus068', 'usedspace': 1L, 'id': 238L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 167L, 'totalspace': 1L, 'hostname': 'locus069', 'usedspace': 1L, 'id': 239L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 168L, 'totalspace': 1L, 'hostname': 'locus070', 'usedspace': 1L, 'id': 240L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 169L, 'totalspace': 1L, 'hostname': 'locus071', 'usedspace': 1L, 'id': 241L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 170L, 'totalspace': 1L, 'hostname': 'locus072', 'usedspace': 1L, 'id': 242L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 171L, 'totalspace': 1L, 'hostname': 'locus073', 'usedspace': 1L, 'id': 243L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 172L, 'totalspace': 1L, 'hostname': 'locus074', 'usedspace': 1L, 'id': 244L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 173L, 'totalspace': 1L, 'hostname': 'locus075', 'usedspace': 1L, 'id': 245L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 174L, 'totalspace': 1L, 'hostname': 'locus076', 'usedspace': 1L, 'id': 246L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 175L, 'totalspace': 1L, 'hostname': 'locus077', 'usedspace': 1L, 'id': 247L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 176L, 'totalspace': 1L, 'hostname': 'locus078', 'usedspace': 1L, 'id': 248L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 177L, 'totalspace': 1L, 'hostname': 'locus079', 'usedspace': 1L, 'id': 249L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 178L, 'totalspace': 1L, 'hostname': 'locus080', 'usedspace': 1L, 'id': 250L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 179L, 'totalspace': 1L, 'hostname': 'locus081', 'usedspace': 1L, 'id': 251L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 180L, 'totalspace': 1L, 'hostname': 'locus082', 'usedspace': 1L, 'id': 252L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 181L, 'totalspace': 1L, 'hostname': 'locus083', 'usedspace': 1L, 'id': 253L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 182L, 'totalspace': 1L, 'hostname': 'locus084', 'usedspace': 1L, 'id': 254L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 183L, 'totalspace': 1L, 'hostname': 'locus085', 'usedspace': 1L, 'id': 255L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 184L, 'totalspace': 1L, 'hostname': 'locus086', 'usedspace': 1L, 'id': 256L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 185L, 'totalspace': 1L, 'hostname': 'locus087', 'usedspace': 1L, 'id': 257L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 186L, 'totalspace': 1L, 'hostname': 'locus088', 'usedspace': 1L, 'id': 258L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 187L, 'totalspace': 1L, 'hostname': 'locus089', 'usedspace': 1L, 'id': 259L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 188L, 'totalspace': 1L, 'hostname': 'locus090', 'usedspace': 1L, 'id': 260L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 189L, 'totalspace': 1L, 'hostname': 'locus091', 'usedspace': 1L, 'id': 261L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 190L, 'totalspace': 1L, 'hostname': 'locus092', 'usedspace': 1L, 'id': 262L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 191L, 'totalspace': 1L, 'hostname': 'locus093', 'usedspace': 1L, 'id': 263L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 192L, 'totalspace': 1L, 'hostname': 'locus094', 'usedspace': 1L, 'id': 264L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 193L, 'totalspace': 1L, 'hostname': 'locus095', 'usedspace': 1L, 'id': 265L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 194L, 'totalspace': 1L, 'hostname': 'locus096', 'usedspace': 1L, 'id': 266L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 195L, 'totalspace': 1L, 'hostname': 'locus097', 'usedspace': 1L, 'id': 267L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 196L, 'totalspace': 1L, 'hostname': 'locus098', 'usedspace': 1L, 'id': 268L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 197L, 'totalspace': 1L, 'hostname': 'locus099', 'usedspace': 1L, 'id': 269L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 198L, 'totalspace': 1L, 'hostname': 'locus100', 'usedspace': 1L, 'id': 270L, 'groupid': 3L, 'claimedspace': 0L, 'path': '/data', 'statusid': 0L}, {'hostid': 199L, 'totalspace': 702716L, 'hostname': 'lustre001', 'usedspace': 23084L, 'id': 271L, 'groupid': 4L, 'claimedspace': 0L, 'path': '/lustre', 'statusid': 1L}]
        mock.getIngestJobs.return_value=[{'username': 'veen', 'description': 'Job: 615122 Start: 2015-12-14T13:03:22 Last update: 2015-12-14T22:12:03 User: veen State: running Name: LOTAAS-2015-12-03 Project: LT5_004 Location: Lofar Storage (SARA)', 'start': '2015-12-14T13:03:22', 'update': '2015-12-14T22:12:03', 'project': 'LT5_004', 'nr_files': 83L, 'state': 'running', 'location': 'Lofar Storage (SARA)', 'obsid': 'L413584', 'id': 615122L, 'name': 'LOTAAS-2015-12-03'}, {'username': 'veen', 'description': 'Job: 615122 Start: 2015-12-14T13:03:22 Last update: 2015-12-14T22:12:03 User: veen State: running Name: LOTAAS-2015-12-03 Project: LT5_004 Location: Lofar Storage (SARA)', 'start': '2015-12-14T13:03:22', 'update': '2015-12-14T22:12:03', 'project': 'LT5_004', 'nr_files': 224L, 'state': 'running', 'location': 'Lofar Storage (SARA)', 'obsid': 'L413588', 'id': 615122L, 'name': 'LOTAAS-2015-12-03'}, {'username': 'veen', 'description': 'Job: 615122 Start: 2015-12-14T13:03:22 Last update: 2015-12-14T22:12:03 User: veen State: running Name: LOTAAS-2015-12-03 Project: LT5_004 Location: Lofar Storage (SARA)', 'start': '2015-12-14T13:03:22', 'update': '2015-12-14T22:12:03', 'project': 'LT5_004', 'nr_files': 224L, 'state': 'running', 'location': 'Lofar Storage (SARA)', 'obsid': 'L413592', 'id': 615122L, 'name': 'LOTAAS-2015-12-03'}, {'username': 'eorru', 'description': 'Job: 622649 Start: 2015-12-11T15:54:56 Last update: 2015-12-12T01:10:22 User: eorru State: running Name: LBS-20151210 Project: LC3_007 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-11T15:54:56', 'update': '2015-12-12T01:10:22', 'project': 'LC3_007', 'nr_files': 1L, 'state': 'running', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L418167', 'id': 622649L, 'name': 'LBS-20151210'}, {'username': 'veen', 'description': 'Job: 614942 Start: 2015-12-14T13:15:47 Last update: 2015-12-14T13:15:47 User: veen State: scheduled Name: LOTAAS-2015-12-02 Project: LT5_004 Location: Lofar Storage (SARA)', 'start': '2015-12-14T13:15:47', 'update': '2015-12-14T13:15:47', 'project': 'LT5_004', 'nr_files': 224L, 'state': 'scheduled', 'location': 'Lofar Storage (SARA)', 'obsid': 'L412954', 'id': 614942L, 'name': 'LOTAAS-2015-12-02'}, {'username': 'donker', 'description': 'Job: 621598 Start: 2015-12-14T13:25:59 Last update: 2015-12-14T13:25:59 User: donker State: scheduled Name: P202+60P184+57 Project: LT5_007 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-14T13:25:59', 'update': '2015-12-14T13:25:59', 'project': 'LT5_007', 'nr_files': 31L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L414813', 'id': 621598L, 'name': 'P202+60P184+57'}, {'username': 'donker', 'description': 'Job: 621598 Start: 2015-12-14T13:25:59 Last update: 2015-12-14T13:25:59 User: donker State: scheduled Name: P202+60P184+57 Project: LT5_007 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-14T13:25:59', 'update': '2015-12-14T13:25:59', 'project': 'LT5_007', 'nr_files': 10L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L414817', 'id': 621598L, 'name': 'P202+60P184+57'}, {'username': 'raf93', 'description': 'Job: 622910 Start: 2015-12-15T08:39:09 Last update: 2015-12-15T08:39:09 User: raf93 State: scheduled Name: December 2015 repeat Project: LT5_003 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-15T08:39:09', 'update': '2015-12-15T08:39:09', 'project': 'LT5_003', 'nr_files': 21L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L414795', 'id': 622910L, 'name': 'December 2015 repeat'}, {'username': 'raf93', 'description': 'Job: 622911 Start: 2015-12-15T08:40:29 Last update: 2015-12-15T08:40:29 User: raf93 State: scheduled Name: Exploratory-11-06 Project: LC4_004 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-15T08:40:29', 'update': '2015-12-15T08:40:29', 'project': 'LC4_004', 'nr_files': 21L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L403274', 'id': 622911L, 'name': 'Exploratory-11-06'}, {'username': 'raf93', 'description': 'Job: 622911 Start: 2015-12-15T08:40:29 Last update: 2015-12-15T08:40:29 User: raf93 State: scheduled Name: Exploratory-11-06 Project: LC4_004 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-15T08:40:29', 'update': '2015-12-15T08:40:29', 'project': 'LC4_004', 'nr_files': 21L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L413773', 'id': 622911L, 'name': 'Exploratory-11-06'}, {'username': 'unknown', 'description': 'MoM Job 6590 not found', 'start': 'unknown', 'update': 'unknown', 'project': 'unknown', 'nr_files': 1L, 'state': 'unknown', 'location': 'unknown', 'obsid': 'LOFAR', 'id': 6590L, 'name': 'unknown'}, {'username': 'unknown', 'description': 'MoM Job 6591 not found', 'start': 'unknown', 'update': 'unknown', 'project': 'unknown', 'nr_files': 1L, 'state': 'unknown', 'location': 'unknown', 'obsid': 'LOFAR', 'id': 6591L, 'name': 'unknown'}]
        mock.getIngestMain.return_value=[{'length': 862L}]
        retvalues_getstatenames      = {'1': 'Active', '0': 'Inactive'}
        retvalues_getactivegroupnames= {'1': 'computenodes', '0': 'storagenodes', '3': 'locusnodes', '2': 'archivenodes', '4': 'cep4'}
        retvalues_gethostsforgid     = {'groupname': 'cep4', 'nodes': [{'statename': 'Active', 'totalspace': 702716L, 'hostname': 'lustre001', 'usedspace': 23084L, 'groupname': 'cep4', 'claimedspace': 0L, 'path': '/lustre', 'id': 1L}]}
        retvalues_listall            = {'domain': {'storage': [], 'name': 'CEP4'}, 'nodes': {'locus085': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus082': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus083': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lse015': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse014': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'lse017': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'lse016': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'lse011': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse010': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse013': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse012': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'locus028': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus029': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus081': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lse019': {'status': 0L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'lse018': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'locus080': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lce039': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce038': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce037': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce036': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce035': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce034': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce033': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce032': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce031': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce030': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus100': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lse020': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'lse021': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'lse022': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'lse023': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'lse024': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 0L}, 'locus036': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus035': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus034': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus039': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus038': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus020': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lce048': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupidJ': 1L}, 'lce049': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus021': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lce042': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce043': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce040': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce041': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce046': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce047': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce044': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce045': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus023': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus024': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus025': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus026': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lexar002': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 2L}, 'locus027': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus006': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus007': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus004': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus005': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus002': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus003': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus001': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lexar001': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 2L}, 'locus008': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus009': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lce059': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce058': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus088': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus089': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lce051': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce050': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce053': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce052': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce055': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce054': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce057': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce056': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus019': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus018': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus011': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus010': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus013': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus012': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus015': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus014': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus017': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus016': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lce064': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce065': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce066': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce067': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce060': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce061': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce062': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce063': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce068': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce069': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus091': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus090': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus093': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus092': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus095': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus094': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus097': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus096': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus033': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus098': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus068': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus069': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus032': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus064': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus065': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus066': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus067': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus060': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus061': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus062': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus063': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lce072': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce071': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce070': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus079': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus078': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus077': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus076': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus075': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus074': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus073': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus072': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus071': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus070': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lce006': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 1L}, 'lce007': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 1L}, 'lce004': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 1L}, 'lce005': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 1L}, 'lce002': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 1L}, 'lce003': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 1L}, 'lce001': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 1L}, 'lce008': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce009': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus099': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus042': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus043': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus040': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus041': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus046': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus047': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus044': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus045': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus048': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus049': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus022': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lce015': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce014': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce017': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce016': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce011': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce010': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce013': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce012': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce019': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce018': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus055': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus054': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus057': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus056': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus051': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus050': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus053': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus052': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus059': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus058': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'lse006': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse007': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse004': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse005': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse002': {'status': 0L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse003': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse001': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lustre001': {'status': 1L, 'storage': [{'path': '/lustre', 'claimedspace': 0L, 'usedspace': 23084L, 'totalspace': 702716L}], 'groupid': 4L}, 'lse008': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lse009': {'status': 1L, 'storage': [{'path': '/data1', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data2', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data3', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}, {'path': '/data4', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 0L}], 'groupid': 0L}, 'lce020': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce021': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 0L, 'totalspace': 1L}], 'groupid': 1L}, 'lce022': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce023': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce024': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce025': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce026': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce027': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce028': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'lce029': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 1L}, 'locus031': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus086': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus030': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus087': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus037': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}, 'locus084': {'status': 0L, 'storage': [{'path': '/data', 'claimedspace': 0L, 'usedspace': 1L, 'totalspace': 1L}], 'groupid': 3L}}}
        retvalues_countactivehosts   = {'locusnodes': {'Active': 0, 'Inactive': 100}, 'cep4': {'Active': 1, 'Inactive': 0}, 'storagenodes': {'Active': 22, 'Inactive': 2}, 'computenodes': {'Active': 0, 'Inactive': 72}, 'archivenodes': {'Active': 0, 'Inactive': 2}}
        retvalues_getArchivingStatus = {'main': [{'length': 862L}], 'jobs': [{'username': 'veen', 'description': 'Job: 615122 Start: 2015-12-14T13:03:22 Last update: 2015-12-14T22:12:03 User: veen State: running Name: LOTAAS-2015-12-03 Project: LT5_004 Location: Lofar Storage (SARA)', 'start': '2015-12-14T13:03:22', 'update': '2015-12-14T22:12:03', 'project': 'LT5_004', 'nr_files': 83L, 'state': 'running', 'location': 'Lofar Storage (SARA)', 'obsid': 'L413584', 'id': 615122L, 'name': 'LOTAAS-2015-12-03'}, {'username': 'veen', 'description': 'Job: 615122 Start: 2015-12-14T13:03:22 Last update: 2015-12-14T22:12:03 User: veen State: running Name: LOTAAS-2015-12-03 Project: LT5_004 Location: Lofar Storage (SARA)', 'start': '2015-12-14T13:03:22', 'update': '2015-12-14T22:12:03', 'project': 'LT5_004', 'nr_files': 224L, 'state': 'running', 'location': 'Lofar Storage (SARA)', 'obsid': 'L413588', 'id': 615122L, 'name': 'LOTAAS-2015-12-03'}, {'username': 'veen', 'description': 'Job: 615122 Start: 2015-12-14T13:03:22 Last update: 2015-12-14T22:12:03 User: veen State: running Name: LOTAAS-2015-12-03 Project: LT5_004 Location: Lofar Storage (SARA)', 'start': '2015-12-14T13:03:22', 'update': '2015-12-14T22:12:03', 'project': 'LT5_004', 'nr_files': 224L, 'state': 'running', 'location': 'Lofar Storage (SARA)', 'obsid': 'L413592', 'id': 615122L, 'name': 'LOTAAS-2015-12-03'}, {'username': 'eorru', 'description': 'Job: 622649 Start: 2015-12-11T15:54:56 Last update: 2015-12-12T01:10:22 User: eorru State: running Name: LBS-20151210 Project: LC3_007 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-11T15:54:56', 'update': '2015-12-12T01:10:22', 'project': 'LC3_007', 'nr_files': 1L, 'state': 'running', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L418167', 'id': 622649L, 'name': 'LBS-20151210'}, {'username': 'veen', 'description': 'Job: 614942 Start: 2015-12-14T13:15:47 Last update: 2015-12-14T13:15:47 User: veen State: scheduled Name: LOTAAS-2015-12-02 Project: LT5_004 Location: Lofar Storage (SARA)', 'start': '2015-12-14T13:15:47', 'update': '2015-12-14T13:15:47', 'project': 'LT5_004', 'nr_files': 224L, 'state': 'scheduled', 'location': 'Lofar Storage (SARA)', 'obsid': 'L412954', 'id': 614942L, 'name': 'LOTAAS-2015-12-02'}, {'username': 'donker', 'description': 'Job: 621598 Start: 2015-12-14T13:25:59 Last update: 2015-12-14T13:25:59 User: donker State: scheduled Name: P202+60P184+57 Project: LT5_007 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-14T13:25:59', 'update': '2015-12-14T13:25:59', 'project': 'LT5_007', 'nr_files': 31L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L414813', 'id': 621598L, 'name': 'P202+60P184+57'}, {'username': 'donker', 'description': 'Job: 621598 Start: 2015-12-14T13:25:59 Last update: 2015-12-14T13:25:59 User: donker State: scheduled Name: P202+60P184+57 Project: LT5_007 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-14T13:25:59', 'update': '2015-12-14T13:25:59', 'project': 'LT5_007', 'nr_files': 10L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L414817', 'id': 621598L, 'name': 'P202+60P184+57'}, {'username': 'raf93', 'description': 'Job: 622910 Start: 2015-12-15T08:39:09 Last update: 2015-12-15T08:39:09 User: raf93 State: scheduled Name: December 2015 repeat Project: LT5_003 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-15T08:39:09', 'update': '2015-12-15T08:39:09', 'project': 'LT5_003', 'nr_files': 21L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L414795', 'id': 622910L, 'name': 'December 2015 repeat'}, {'username': 'raf93', 'description': 'Job: 622911 Start: 2015-12-15T08:40:29 Last update: 2015-12-15T08:40:29 User: raf93 State: scheduled Name: Exploratory-11-06 Project: LC4_004 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-15T08:40:29', 'update': '2015-12-15T08:40:29', 'project': 'LC4_004', 'nr_files': 21L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L403274', 'id': 622911L, 'name': 'Exploratory-11-06'}, {'username': 'raf93', 'description': 'Job: 622911 Start: 2015-12-15T08:40:29 Last update: 2015-12-15T08:40:29 User: raf93 State: scheduled Name: Exploratory-11-06 Project: LC4_004 Location: Lofar Storage (J\xc3\xbclich)', 'start': '2015-12-15T08:40:29', 'update': '2015-12-15T08:40:29', 'project': 'LC4_004', 'nr_files': 21L, 'state': 'scheduled', 'location': 'Lofar Storage (J\xc3\xbclich)', 'obsid': 'L413773', 'id': 622911L, 'name': 'Exploratory-11-06'}, {'username': 'unknown', 'description': 'MoM Job 6590 not found', 'start': 'unknown', 'update': 'unknown', 'project': 'unknown', 'nr_files': 1L, 'state': 'unknown', 'location': 'unknown', 'obsid': 'LOFAR', 'id': 6590L, 'name': 'unknown'}, {'username': 'unknown', 'description': 'MoM Job 6591 not found', 'start': 'unknown', 'update': 'unknown', 'project': 'unknown', 'nr_files': 1L, 'state': 'unknown', 'location': 'unknown', 'obsid': 'LOFAR', 'id': 6591L, 'name': 'unknown'}]}

        class Test1(unittest.TestCase):
            '''Test'''

            def test(self):
                '''basic test '''
                rpc = SSDBRPC(busname=busname, servicename=servicename)
                self.assertEqual(retvalues_getstatenames      , rpc.getstatenames())
                self.assertEqual(retvalues_getactivegroupnames, rpc.getactivegroupnames())
                self.assertEqual(retvalues_gethostsforgid     , rpc.gethostsforgid("4"))
                #self.assertEqual((retvalues_listall            , rpc.listall())
                tempout=rpc.listall()
                self.assertEqual(retvalues_countactivehosts   , rpc.countactivehosts())
                self.assertEqual(retvalues_getArchivingStatus , rpc.getArchivingStatus())

        # create and run the service
        with createService(busname=busname, servicename=servicename):
            # and run all tests
            unittest.main()
            print "done testing"

finally:
    # cleanup test bus and exit
    if broker:
        broker.delExchange(busname)
    if connection:
        connection.close()
